from protowhat.utils_messaging import get_ord
from protowhat.sct_syntax import link_to_state
from tcs_pythonwhat.checks.check_funcs import (
    check_node,
    check_part,
    check_part_index,
    with_context,
)
from tcs_pythonwhat.test_funcs.utils import fix_format, stringify, call
from tcs_pythonwhat.checks.check_logic import multi
from tcs_pythonwhat.checks.has_funcs import (
    has_equal_part_len,
    has_equal_part,
    has_equal_value,
    has_equal_output,
)
from tcs_pythonwhat.checks.check_has_context import has_context
from functools import partial

# this is done by the chain for v2
# it's only needed when a new state is created and (possibly) used elsewhere
check_node = link_to_state(check_node)
check_part = link_to_state(check_part)
check_part_index = link_to_state(check_part_index)


def test_if_else(state, index=1, test=None, body=None, orelse=None):
    """Test parts of the if statement.

    This test function will allow you to extract parts of a specific if statement and perform a set of tests
    specifically on these parts. A for loop consists of three potential parts: the condition test, :code:`test`,
    which specifies the condition of the if statement, the :code:`body`, which is what's executed if the condition is
    True and a else part, :code:`orelse`, which will be executed if the condition is not True.::

        if 5 == 3:
            print("success")
        else:
            print("fail")

    Has :code:`5 == 3` as the condition test, :code:`print("success")` as the body and :code:`print("fail")` as the else part.

    Args:
      index (int): index of the function call to be checked. Defaults to 1.
      test: this argument holds the part of code that will be ran to check the condition test of the if statement.
        It should be passed as a lambda expression or a function definition. The functions that are ran should
        be other pythonwhat test functions, and they will be tested specifically on only the condition test of
        the if statement.
      body: this argument holds the part of code that will be ran to check the body of the if statement.
        It should be passed as a lambda expression or a function definition. The functions that are ran should
        be other pythonwhat test functions, and they will be tested specifically on only the body of
        the if statement.
      orelse: this argument holds the part of code that will be ran to check the else part of the if statement.
        It should be passed as a lambda expression or a function definition. The functions that are ran should
        be other pythonwhat test functions, and they will be tested specifically on only the else part of
        the if statement.

    :Example:

        Student code::

            a = 12
            if a > 3:
                print('test %d' % a)

        Solution code::

            a = 4
            if a > 3:
                print('test %d' % a)

        SCT::

            test_if_else(1,
                body = test_expression_output(
                        extra_env = { 'a': 5 }
                        incorrect_msg = "Print out the correct things"))

        This SCT will pass as :code:`test_expression_output()` is ran on the body of the if statement and it will output
        the same thing in the solution as in the student code.
    """
    state = check_node(
        state, "if_elses", index - 1, typestr="{{ordinal}} if expression"
    )
    multi(check_part(state, "test", "condition"), test)
    multi(check_part(state, "body", "body"), body)
    multi(check_part(state, "orelse", "else part"), orelse)


def test_for_loop(state, index=1, for_iter=None, body=None, orelse=None):
    """Test parts of the for loop.

    This test function will allow you to extract parts of a specific for loop and perform a set of tests
    specifically on these parts. A for loop consists of two parts: the sequence, `for_iter`, which is the
    values over which are looped, and the `body`. A for loop can have a else part as well, `orelse`, but
    this is almost never used.::

        for i in range(10):
            print(i)

    Has :code:`range(10)` as the sequence and :code:`print(i)` as the body.

    Args:
      index (int): index of the function call to be checked. Defaults to 1.
      for_iter: this argument holds the part of code that will be ran to check the sequence of the for loop.
        It should be passed as a lambda expression or a function. The functions that are ran should
        be other pythonwhat test functions, and they will be tested specifically on only the sequence part of
        the for loop.
      body: this argument holds the part of code that will be ran to check the body of the for loop.
        It should be passed as a lambda expression or a function. The functions that are ran should
        be other pythonwhat test functions, and they will be tested specifically on only the body of
        the for loop.
      orelse: this argument holds the part of code that will be ran to check the else part of the for loop.
        It should be passed as a lambda expression or a function. The functions that are ran should
        be other pythonwhat test functions, and they will be tested specifically on only the else part of
        the for loop.

    :Example:
        Student code::

            for i in range(10):
                print(i)

        Solution code::

            for n in range(10):
                print(n)

        SCT::

            test_for_loop(1,
                for_iter = test_function("range"),
                body = test_expression_output(context_val = [5])

        This SCT will evaluate to True as the function :code:`range` is used in the sequence and the function
        :code:`test_exression_output()` will pass on the body code.
    """
    state = check_node(state, "for_loops", index - 1, "{{ordinal}} for loop")

    multi(check_part(state, "iter", "sequence part"), for_iter)
    multi(check_part(state, "body", "body"), body)
    multi(check_part(state, "orelse", "else part"), orelse)


def test_while_loop(state, index=1, test=None, body=None, orelse=None):
    """Test parts of the while loop.

    This test function will allow you to extract parts of a specific while loop and perform a set of tests
    specifically on these parts. A while loop generally consists of two parts: the condition test, :code:`test`,
    which is the condition that is tested each loop, and the :code:`body`. A for while can have a else part as well,
    :code:`orelse`, but this is almost never used.::

        a = 10
        while a < 5:
            print(a)
            a -= 1

    Has :code:`a < 5` as the condition test and `print(i)` as the body.

    Args:
        index (int): index of the function call to be checked. Defaults to 1.
        test: this argument holds the part of code that will be ran to check the condition test of the while loop.
          It should be passed as a lambda expression or a function definition. The functions that are ran should
          be other pythonwhat test functions, and they will be tested specifically on only the condition test of
          the while loop.
        body: this argument holds the part of code that will be ran to check the body of the while loop.
          It should be passed as a lambda expression or a function definition. The functions that are ran should
          be other pythonwhat test functions, and they will be tested specifically on only the body of
          the while loop.
        orelse: this argument holds the part of code that will be ran to check the else part of the while loop.
          It should be passed as a lambda expression or a function definition. The functions that are ran should
          be other pythonwhat test functions, and they will be tested specifically on only the else part of
          the while loop.

    :Example:

        Student code::

            a = 10
            while a < 5:
                print(a)
                a -= 1

        Solution code::

            a = 20
            while a < 5:
                print(a)
                a -= 1

        SCT::

            test_while_loop(1,
                    test = test_expression_result({"a": 5}),
                    body = test_expression_output({"a": 5}))

      This SCT will evaluate to True as condition test will have thes same result in student
      and solution code and `test_exression_output()` will pass on the body code.
    """
    state = check_node(state, "whiles", index - 1, "{{ordinal}} while loop")
    multi(check_part(state, "test", "condition"), test)
    multi(check_part(state, "body", "body"), body)
    multi(check_part(state, "orelse", "else part"), orelse)


def test_function_definition(
    state,
    name,
    arg_names=True,
    arg_defaults=True,
    body=None,
    results=None,
    outputs=None,
    errors=None,
    not_called_msg=None,
    nb_args_msg=None,
    other_args_msg=None,
    arg_names_msg=None,
    arg_defaults_msg=None,
    wrong_result_msg=None,
    wrong_output_msg=None,
    no_error_msg=None,
    wrong_error_msg=None,
):
    """Test a function definition.

    This function helps you test a function definition. Generally four things can be tested:
        1) The argument names of the function (including if the correct defaults are used)
        2) The body of the functions (does it output correctly, are the correct functions used)
        3) The return value with a certain input
        4) The output value with a certain input
        5) Whether certain inputs generate an error and what type of error

    Custom feedback messages can be set for all these parts, default messages are generated
    automatically if none are set.

    Args:
        name (str): the name of the function definition to be tested.
        arg_names (bool): if True, the argument names will be tested, if False they won't be tested. Defaults
            to True.
        arg_defaults (bool): if True, the default values of the arguments will be tested, if False they won't
            be tested. Defaults to True.
        body: this arguments holds the part of the code that will be ran to check the body of the function
            definition. It should be passed as a lambda expression or a function. The functions that are
            ran should be other pythonwhat test functions, and they will be tested specifically on only the
            body of the for loop. Defaults to None.
        results (list(list)): a list of lists representing arguments that should be passed to the defined
            function. These arguments are passed to the function in the student environment and the solution
            environment, the results (what's returned) are compared.
        outputs (list(list)): a list of lists representing arguments that should be passed to the defined
            function. These arguments are passed to the function in the student environment and the solution
            environment, the outpus are compared.
        errors (list(list)): a list of lists representing arguments that should be passed to the defined
            function. These arguments are passed to the function in the student environment and the solution
            environment, the errors they generate are compared.
        not_called_msg (str): message if the function is not defined.
        nb_args_msg (str): message if the number of arguments do not matched.
        arg_names_msg (str): message if the argument names do not match.
        arg_defaults_msg (str): message if the argument default values do not match.
        wrong_result_msg (str): message if one of the tested function calls' result did not match.
        wrong_output_msg (str): message if one of the tested functions calls' output did not match.
        no_error_msg (str): message if one of the tested function calls' result did not generate an error.
        wrong_error_msg (str): message if the error that one of the tested function calls generated did not match.

    :Example:

        Student code::

            def shout( word, times = 3):
                shout_word = not_word + '???'
                print( shout_word )
                return word * times

        Solution code::

            def shout( word = 'help', times = 3 ):
                shout_word = word + '!!!'
                print( shout_word )
                return word * times

        SCT::

            test_function_definition('shout')                          # fail
            test_function_definition('shout', arg_defaults = False)    # pass
            test_function_definition('shout', arg_defaults = False,    # fail
                                            outputs = [('help')])

            test_function_definition('shout', arg_defaults = False,    # pass
                                            results = [('help', 2)])

            test_function_definition('shout', args_defaults = False    # pass
                    body = test_function('print', args = []]))
    """

    # what the function will be referred to as
    child = check_node(state, "function_defs", name, "definition of `{{index}}()`")

    test_args(
        child, arg_names, arg_defaults, nb_args_msg, arg_names_msg, arg_defaults_msg
    )

    multi(check_part(child, "body", ""), body)

    # Test function calls -----------------------------------------------------

    for el in results or []:
        el = fix_format(el)
        call(
            child,
            el,
            "value",
            incorrect_msg=wrong_result_msg,
            error_msg=wrong_result_msg,
            argstr="`{}{}`".format(name, stringify(el)),
        )

    for el in outputs or []:
        el = fix_format(el)
        call(
            child,
            el,
            "output",
            incorrect_msg=wrong_output_msg,
            error_msg=wrong_output_msg,
            argstr="`{}{}`".format(name, stringify(el)),
        )

    for el in errors or []:
        el = fix_format(el)
        call(
            child,
            el,
            "error",
            incorrect_msg=wrong_error_msg,
            error_msg=no_error_msg,
            argstr="`{}{}`".format(name, stringify(el)),
        )


def test_args(
    state, arg_names, arg_defaults, nb_args_msg, arg_names_msg, arg_defaults_msg
):

    MSG_NUM_ARGS = "You should define {{parent[typestr]}} with {{sol_len}} arguments, instead got {{stu_len}}."
    MSG_BAD_ARG_NAME = "The {{parent[ordinal]}} {{parent[part]}} should be called `{{sol_part[name]}}`, instead got `{{stu_part[name]}}`."
    MSG_BAD_DEFAULT = (
        "The {{parent[part]}} `{{stu_part[name]}}` should have no default."
    )
    MSG_INC_DEFAULT = (
        "The {{parent[part]}} `{{stu_part[name]}}` does not have the correct default."
    )

    MSG_NO_VARARG = "Вы указали аргумент для принятия аргумента '*' и назвали его `{{sol_part['*args'][name]}}`?"
    MSG_NO_KWARGS = "Вы указали аргумент для принятия аргумента '**' и назвали его `{{sol_part['**kwargs'][name]}}`?"
    MSG_VARARG_NAME = "Вы указали аргумент для принятия аргумента '*' и назвали его `{{sol_part[name]}}`?"
    MSG_KWARG_NAME = "Вы указали аргумент для принятия аргумента '** и назвали его `{{sol_part[name]}}`?"

    if arg_names or arg_defaults:
        # test number of args
        has_equal_part_len(state, "_spec1_args", nb_args_msg or MSG_NUM_ARGS)

        # iterate over each arg, testing name and default
        for ii in range(len(state.solution_parts["_spec1_args"])):
            # get argument state
            arg_state = check_part_index(
                state, "_spec1_args", ii, "argument", "NO MISSING MSG"
            )
            # test exact name
            has_equal_part(arg_state, "name", arg_names_msg or MSG_BAD_ARG_NAME)

            if arg_defaults:
                # test whether is default
                has_equal_part(
                    arg_state, "is_default", arg_defaults_msg or MSG_BAD_DEFAULT
                )
                # test default value, use if to prevent running a process no default
                if arg_state.solution_parts["is_default"]:
                    has_equal_value(
                        arg_state,
                        incorrect_msg=arg_defaults_msg or MSG_INC_DEFAULT,
                        append=True,
                    )

        # test *args and **kwargs
        if state.solution_parts["*args"]:
            vararg = check_part(state, "*args", "", missing_msg=MSG_NO_VARARG)
            has_equal_part(vararg, "name", MSG_VARARG_NAME)

        if state.solution_parts["**kwargs"]:
            kwarg = check_part(state, "**kwargs", "", missing_msg=MSG_NO_KWARGS)
            has_equal_part(kwarg, "name", MSG_KWARG_NAME)


def test_expression_result(
    state,
    extra_env=None,
    context_vals=None,
    incorrect_msg=None,
    expr_code=None,
    pre_code=None,
    error_msg=None,
    **kwargs
):
    has_equal_value(
        state,
        incorrect_msg=incorrect_msg,
        error_msg=error_msg,
        extra_env=extra_env,
        context_vals=context_vals,
        expr_code=expr_code,
        pre_code=pre_code,
        **kwargs
    )


def test_expression_output(
    state,
    extra_env=None,
    context_vals=None,
    incorrect_msg=None,
    eq_condition="equal",
    expr_code=None,
    pre_code=None,
    **kwargs
):
    has_equal_output(
        state,
        incorrect_msg=incorrect_msg,
        extra_env=extra_env,
        context_vals=context_vals,
        expr_code=expr_code,
        pre_code=pre_code,
        **kwargs
    )


def test_object_after_expression(
    state,
    name,
    extra_env=None,
    context_vals=None,
    undefined_msg=None,
    incorrect_msg=None,
    expr_code=None,
    pre_code=None,
    **kwargs
):
    state.highlight = (
        state.ast_dispatcher.find("object_assignments", state.student_ast)
        .get(name, {})
        .get("highlight")
    )
    has_equal_value(
        state,
        incorrect_msg=incorrect_msg,
        error_msg=undefined_msg,
        undefined_msg=undefined_msg,
        extra_env=extra_env,
        context_vals=context_vals,
        pre_code=pre_code,
        name=name,
        expr_code=expr_code,
        **kwargs
    )


def test_with(
    state,
    index,
    context_vals=False,  # whether to check number of context vals
    context_tests=None,  # check on context expressions
    body=None,
    undefined_msg=None,
    context_vals_len_msg=None,
    context_vals_msg=None,
):
    """Test a with statement.
with open_file('...') as bla:

    [ open_file('...').__enter__() ]


with open_file('...') as file:
    [ ]

    """

    MSG_NUM_CTXT = "Убедитесь, что используется правильное количество контекстных переменной. Кажется, вы определили слишком много."
    MSG_NUM_CTXT2 = "Убедитесь, что используется правильное количество контекстных переменных. Похоже, вы определили слишком мало."
    MSG_CTXT_NAMES = "Убедитесь, что вы используете правильные имена контекстных переменных. Ожидалось `{{sol_vars}}`, но было получено `{{stu_vars}}`."

    check_with = partial(
        check_node, state, "withs", index - 1, "{{ordinal}} `with` statement"
    )

    child = check_with()
    child2 = check_with()

    if context_vals:
        # test context var names ----
        has_context(
            child, incorrect_msg=context_vals_msg or MSG_CTXT_NAMES, exact_names=True
        )

        # test num context vars ----
        has_equal_part_len(child, "context", MSG_NUM_CTXT)

    # Context sub tests ----
    if context_tests and not isinstance(context_tests, list):
        context_tests = [context_tests]

    for i, context_test in enumerate(context_tests or []):
        # partial the substate check, because the function uses two prepended messages
        def check_context(state):
            return check_part_index(
                state,
                "context",
                i,
                "%s context" % get_ord(i + 1),
                missing_msg=MSG_NUM_CTXT2,
            )

        check_context(child)  # test exist

        ctxt_state = check_context(child2)  # sub tests
        multi(ctxt_state, context_test)

    # Body sub tests ----
    if body is not None:
        body_state = check_part(child2, "body", "body")

        with_context(body_state, body, child=child)


def test_list_comp(
    state,
    index=1,
    not_called_msg=None,
    comp_iter=None,
    iter_vars_names=False,
    incorrect_iter_vars_msg=None,
    body=None,
    ifs=None,
    insufficient_ifs_msg=None,
):
    """Test list comprehension."""
    kwargs = locals().copy()
    kwargs["typestr"] = "{{ordinal}} list comprehension"
    kwargs["comptype"] = "list_comps"
    test_comp(**kwargs)


def test_comp(
    state,
    typestr,
    comptype,
    index,
    iter_vars_names,
    not_called_msg,
    insufficient_ifs_msg,
    incorrect_iter_vars_msg,
    comp_iter,
    ifs,
    key=None,
    body=None,
    value=None,
    rep=None,
):

    MSG_INCORRECT_ITER_VARS = "Вы использовали правильные переменные итератора?"
    MSG_INCORRECT_NUM_ITER_VARS = "Вы использовали переменную итератора {{num_vars}}?"
    MSG_INSUFFICIENT_IFS = "Вы использовали {{sol_len}} if'ов?"

    # make sure other messages are set to default if None
    if insufficient_ifs_msg is None:
        insufficient_ifs_msg = MSG_INSUFFICIENT_IFS

    # get comprehension
    child = check_node(state, comptype, index - 1, typestr, missing_msg=not_called_msg)

    # test comprehension iter and its variable names (or number of variables)
    if comp_iter:
        multi(check_part(child, "iter", "iterable part"), comp_iter)

    # test iterator variables
    default_msg = (
        MSG_INCORRECT_ITER_VARS if iter_vars_names else MSG_INCORRECT_NUM_ITER_VARS
    )
    has_context(child, incorrect_iter_vars_msg or default_msg, iter_vars_names)

    # test the main expressions.
    if body:
        multi(check_part(child, "body", "body"), body)  # list and gen comp
    if key:
        multi(check_part(child, "key", "key part"), key)  # dict comp
    if value:
        multi(check_part(child, "value", "value part"), value)  # ""

    # test a list of ifs. each entry corresponds to a filter in the comprehension.
    for i, if_test in enumerate(ifs or []):
        # test that ifs are same length
        has_equal_part_len(child, "ifs", insufficient_ifs_msg)
        # test individual ifs
        multi(check_part_index(child, "ifs", i, get_ord(i + 1) + " if"), if_test)



import ast
from tcs_pythonwhat.utils_ast import wrap_in_module
from collections.abc import Sequence, Mapping
from collections import OrderedDict
from contextlib import ExitStack
from functools import wraps

"""
This file handles the parsing of the student and solution code. Generally, an abstract syntax tree
is built from the code string and this tree is passed to several visitors to create the data
structures that are used by the tests.

For information about how the ast package works, I refer to its documentation:
    https://docs.python.org/2/library/ast.html
as well as some extra documentation:
    https://greentreesnakes.readthedocs.org/en/latest/
"""


class EmptyTargetVar:
    pass


class TargetVars(Mapping):
    """Immutable ordered mapping from target variables to their values."""

    EMPTY = EmptyTargetVar()

    def __init__(self, target_vars=tuple(), is_empty=True):
        if is_empty:
            target_vars = [(v, self.EMPTY) for v in target_vars]

        self._od = OrderedDict(target_vars)

    # getitem, len, iter wrap OrderedDict behavior
    def __getitem__(self, k):
        return self._od.__getitem__(k)

    def __len__(self):
        return self._od.__len__()

    def __iter__(self):
        return self._od.__iter__()

    def update(self, *args, **kwargs):
        cpy = self.copy()
        cpy._od.update(*args, **kwargs)
        return cpy

    def copy(self):
        return self.__class__(self._od)

    def __str__(self):
        """Format target vars for printing"""
        if len(self) > 1:
            return "({})".format(", ".join(self._od.keys()))
        else:
            return "".join(self._od.keys())

    def defined_items(self):
        """Return copy of instance, omitting entries that are EMPTY"""
        return self.__class__(
            [(k, v) for k, v in self.items() if v is not self.EMPTY], is_empty=False
        )


class IndexedDict(Mapping):
    """Wrapper around OrderedDict that allows access via item position or key"""

    def __init__(self, *args, **kwargs):
        self._od = OrderedDict(*args, **kwargs)

    def __getitem__(self, k):
        try:
            return list(self._od.values())[k]
        except TypeError:
            return self._od[k]

    def __len__(self):
        return self._od.__len__()

    def __iter__(self):
        return self._od.__iter__()


class Parser(ast.NodeVisitor):
    """Basic parser.

    The basic Parser, should not be used directly, but to inherit from. The Parser itself inherits
    from ast.Nodevisitor, which is a helper class to go through the abstract syntax tree objects.

    In the basic version, each node in the Module body will be visited. Expression bodies will be
    visited as well. In this standard parser, all other nodes are ignored.
    """

    def visit_Module(self, node):
        """
        This function is called when a Module node is encountered when traversing the tree.

        Args:
            node (ast.Module): The node which is visited.
        """
        for line in node.body:
            # We only want to visit the module nodes on a first level. Going deeper
            # should be handeled by the specific parsers and the test function, as
            # nesting requires the State to generate a subtree. The Parser object
            # does not know about the State object.
            self.visit(line)

    def visit_Expr(self, node):
        self.visit(node.value)

    def visit_Expression(self, node):
        self.visit(node.body)

    def generic_visit(self, node):
        """
        This function is called when all other nodes are encountered when traversing the tree.
        When inheriting form this standard parser, this function will make the parser ignore
        all nodes that are not relevant.

        Args:
            node (ast.Node): The node which is visited.
        """
        pass  # This ignore is necessary to keep the parser at base level, also look comment above in
        # the visit_Module function body.

    def visit_each(self, lst):
        for el in lst:
            self.visit(el)

    @staticmethod
    def get_target_vars(target):
        get_id = lambda n: n.id if not isinstance(n, ast.Starred) else n.value.id
        if isinstance(target, (ast.Name, ast.Starred)):
            tv = [get_id(target)]
        elif isinstance(target, ast.Tuple):
            tv = [get_id(node) for node in target.elts]
        else:
            tv = []

        return TargetVars(tv)

    @staticmethod
    def get_arg(el):
        if el is None:
            return None
        else:
            return el.arg

    @staticmethod
    def get_arg_tuples(arguments, defaults):
        arguments = [arg.arg for arg in arguments]
        defaults = [None] * (len(arguments) - len(defaults)) + defaults
        return list(zip(arguments, defaults))

    @staticmethod
    def get_arg_parts(arguments, defaults, type):
        # only difference is that it doesn't pull out arg.arg, so we can
        # use all the information on the arg node down the road
        match_def = [None] * (len(arguments) - len(defaults)) + defaults
        part_list = []
        for _arg, _def in zip(arguments, match_def):
            part_list.append(Parser.get_arg_part(_arg, _def))
        return part_list

    @staticmethod
    def get_arg_part(_arg, _def, type=None):
        # type is arg, kwonly, kwarg, vararg
        if not _arg:
            return None

        # part uses default highlighting, so will highlight "node" entry
        return {
            "node": _def or _arg,
            "arg": _arg,
            # TODO: need to fill out
            "type": type,
            "is_default": True if _def else False,
            "name": _arg.arg,
            "annotation": _arg.annotation,
        }


# class OperatorParser(Parser):
#     """Find operations.

#     A parser which inherits from the basic parser to find binary operators.

#     Attributes:
#         out (list(tuple(num, ast.BinOp, list(str)))): A list of tuples containing the linenumber, node and list of used binary operations.
#         level (num): A number representing the level at which the parser is parsing.
#         used (list(str)): The operators that are used in the BinOp that we're handling.
#     """


#     # All possible operations and their sign
#     O_MAP = {}
#     O_MAP['Add'] = '+'
#     O_MAP['Sub'] = '-'
#     O_MAP['Mult'] = '*'
#     O_MAP['Div'] = '/'
#     O_MAP['Mod'] = '%'
#     O_MAP['Pow'] = '**'
#     O_MAP['LShift'] = '<<'
#     O_MAP['RShift'] = '>>'
#     O_MAP['BitOr'] = '|'
#     O_MAP['BitXor'] = '^'
#     O_MAP['BitAnd'] = '&'
#     O_MAP['FloorDiv'] = '//'

#     def __init__(self):
#         """
#         Initialize the parser and its attributes.
#         """
#         self.out = []
#         self.level = 0
#         self.used = []

#     def visit_Expr(self, node):
#         self.visit(node.value)

#     def visit_Call(self, node):
#         for arg in node.args:
#             self.visit(arg)

#     def visit_Assign(self, node):
#         self.visit(node.value)

#     def visit_Num(self, node):
#         if not self.level:
#             self.out.append((  # A number can be seen as a operator on base level.
#                 node,          # When student is asked to use operators but just puts in a number instead,
#                 self.used))    # this will help creating a consistent feedback message.

#     def visit_UnaryOp(self, node):
#         self.visit(node.operand)  # Unary operations, like '-', should not be added, but they should be
#         # looked into. They can contain more binary operations. This is important
#         # during the nesting process.

#     def visit_BinOp(self, node):
#         self.used.append(OperatorParser.O_MAP[type(node.op).__name__])
#         self.level = self.level + 1
#         # Nest to other operations, but increase the level. We only
#         self.visit(node.left)
#         # want to now which operations are used at a deeper level, but
#         self.visit(node.right)
#         self.level = self.level - 1  # we don't need all the explicit nodes.

#         if not self.level:          # We should only add the binary operations of the base level,
#             self.out.append((       # information about nested operations is included in the used list.
#                 node,
#                 self.used))
#             self.used = []


class ImportParser(Parser):
    """Find import statement.

    A parser which inherits from the basic parser to find package imports.
    """

    def __init__(self):
        self.out = {}

    def visit_Import(self, node):
        for imp in node.names:
            self.out[imp.name] = imp.asname

    def visit_ImportFrom(self, node):
        for imp in node.names:
            self.out[node.module + "." + imp.name] = imp.asname


class FunctionParser(Parser):
    """Find function calls.

    A parser which inherits from the basic parser to find function calls.
    Function calls inside control structures are not found, nesting function calls are.
    """

    def __init__(self):
        self.gen_name = ""
        self.raw_name = ""
        self.mappings = {}
        self.out = {}
        self.call_lookup_active = False

    def visit_BinOp(self, node):
        self.visit(node.left)
        self.visit(node.right)

    def visit_Assign(self, node):
        self.visit(node.value)

    def visit_AugAssign(self, node):
        self.visit(node.value)

    def visit_Compare(self, node):
        self.visit(node.left)
        self.visit_each(node.comparators)

    def visit_UnaryOp(self, node):
        self.visit(node.operand)

    def visit_Import(self, node):
        for imp in node.names:
            if imp.asname is not None:
                self.mappings[imp.asname] = imp.name
            else:
                pass  # e.g. numpy import as numpy, so no action needed.

    def visit_ImportFrom(self, node):
        for imp in node.names:
            self.mappings[imp.asname or imp.name] = node.module + "." + imp.name

    def visit_Expr(self, node):
        self.visit(node.value)

    def visit_List(self, node):
        [self.visit(el) for el in node.elts]

    def visit_Dict(self, node):
        [self.visit(el) for el in node.values]

    def visit_Call(self, node):
        if self.call_lookup_active:
            self.visit(node.func)
        else:
            self.call_lookup_active = True
            self.visit(
                node.func
            )  # Need to visit func to start recording the current function name.

            if self.gen_name:
                if self.gen_name not in self.out:
                    self.out[self.gen_name] = []

                self.out[self.gen_name].append(self.get_call_part(node))
                # self.out[self.current].append((node, node.args, node.keywords))

            self.gen_name = self.raw_name = ""
            self.call_lookup_active = False

            # dive deeper in func, args and keywords
            self.visit(node.func)

            for arg in node.args:
                self.visit(arg)

            for key in node.keywords:
                self.visit(key.value)

    def visit_Attribute(self, node):
        self.visit(node.value)  # Go deeper for the package/module names!
        self.gen_name += "." + node.attr  # Add the function name
        self.raw_name += "." + node.attr

    def visit_Subscript(self, node):
        # jump over subscripts for the sake of method calls
        self.visit(node.value)

    def visit_Name(self, node):
        self.gen_name = self.mappings.get(node.id) or node.id
        self.raw_name = node.id

    def get_call_part(self, node):
        args = [self.get_pos_arg_part(n, ii) for ii, n in enumerate(node.args)]
        keywords = [self.get_kw_arg_part(n) for n in node.keywords]
        return {
            "node": node,
            # TODO: right now, args and keywords can be indexed by pos or name.
            #       Note that a pos args name is its position.
            #       Problems will arise if SCT tests a position, but the submission
            #       has too few positional arguments, since it will then grab a kw arg :(
            #       This is not necessarily a bad thing, but instructors would need to be
            #       Careful deciding when to test a pos arg, and when to test using kw.
            #       Could use check_pos_args with pos_args entry below to solve.
            "args": IndexedDict((n["name"], n) for n in [*args, *keywords]),
            #'pos_args': args,
            #'keywords': keywords,
            "name": self.raw_name,
        }

    @staticmethod
    def get_pos_arg_part(arg, indx_pos):
        is_star = isinstance(arg, ast.Starred)
        return {
            "node": arg if not is_star else arg.value,
            "highlight": arg,
            "type": "argument",
            "is_starred": is_star,
            "name": indx_pos,
        }

    @staticmethod
    def get_kw_arg_part(arg):
        is_kwarg = arg.arg is None
        return {
            "node": arg.value,
            "highlight": arg,
            "type": "keyword",
            "is_kwarg": is_kwarg,
            "name": arg.arg,
        }


class ObjectAccessParser(FunctionParser):
    """Find object accesses

    A parser which inherits from the FunctionParser to find object accesses.
    """

    def __init__(self):
        super().__init__()
        self.out = []

    def visit_Call(self, node):
        for arg in node.args:
            self.visit(arg)

        for key in node.keywords:
            self.visit(key.value)

    def visit_List(self, node):
        for el in node.elts:
            self.visit(el)

    def visit_Tuple(self, node):
        for el in node.elts:
            self.visit(el)

    def visit_Attribute(self, node):
        # if already a chain, prepend, else initialize self.current
        self.gen_name = node.attr + "." + self.gen_name if self.gen_name else node.attr
        self.raw_name = node.attr + "." + self.raw_name if self.raw_name else node.attr
        self.visit(node.value)

    def visit_Name(self, node):
        # if name refers to an import, replace
        prefix = self.mappings.get(node.id) or node.id

        self.gen_name = prefix + "." + self.gen_name if self.gen_name else prefix
        self.raw_name = node.id + "." + self.raw_name if self.raw_name else node.id

        self.out.append(self.gen_name)
        self.gen_name = self.raw_name = ""


class ObjectAssignmentParser(Parser):
    """Find object assignmnts

    A parser which inherits from the basic parser to find object assignments.
    All assignments at top-level, as well as in if, while, for and with statements are found.
    """

    def __init__(self):
        self.out = {}
        self.active_assignment = None

    def visit_Name(self, node):
        if self.active_assignment is not None:
            if node.id not in self.out:
                self.out[node.id] = self.get_part(node, self.active_assignment)
            else:
                self.out[node.id]["highlight"] = None
            self.active_assignment = None

    def visit_Attribute(self, node):
        self.visit(node.value)

    def visit_Assign(self, node):
        self.active_assignment = node
        self.visit_each(node.targets)

    def visit_AugAssign(self, node):
        self.active_assignment = node
        self.visit(node.target)

    def visit_If(self, node):
        self.visit_each(node.body)
        self.visit_each(node.orelse)

    def visit_While(self, node):
        self.visit_each(node.body)
        self.visit_each(node.orelse)

    def visit_For(self, node):
        self.visit_each(node.body)
        self.visit_each(node.orelse)

    def visit_With(self, node):
        self.visit_each(node.body)

    def visit_Try(self, node):
        self.visit_each(node.body)
        self.visit_each(node.finalbody)

    @staticmethod
    def get_part(name_node, ass_node=None):
        # either name node or simply str or name itself
        name = getattr(name_node, "id", name_node)
        load_name = ast.Name(id=name, ctx=ast.Load())
        ast.fix_missing_locations(load_name)
        return {"name": name, "node": load_name, "highlight": ass_node or name_node}


class IfParser(Parser):
    """Find if structures.

    A parser which inherits from the basic parser to find if structures.
    Only 'top-level' if structures will be found!
    """

    def __init__(self):
        self.out = []

    def visit_If(self, node):
        self.out.append(
            {"node": node, "test": node.test, "body": node.body, "orelse": node.orelse}
        )


class IfExpParser(IfParser):
    """Find if structures.

    A parser which inherits from the basic parser to find inline if structures.
    Only 'top-level' if structures will be found!
    """

    def visit_If(self, node):
        return

    def visit_IfExp(self, node):
        super().visit_If(node)

    def visit_BinOp(self, node):
        self.visit(node.left)
        self.visit(node.right)

    def visit_Assign(self, node):
        self.visit(node.value)

    def visit_AugAssign(self, node):
        self.visit(node.value)

    def visit_Compare(self, node):
        self.visit_each(node.comparators)

    def visit_UnaryOp(self, node):
        self.visit(node.operand)


class WhileParser(Parser):
    """Find while structures.

    A parser which inherits from the basic parser to find while structures.
    Only 'top-level' while structures will be found!
    """

    def __init__(self):
        self.out = []

    def visit_While(self, node):
        self.out.append(
            {"node": node, "test": node.test, "body": node.body, "orelse": node.orelse}
        )


class ForParser(Parser):
    """Find for structures.

    A parser which inherits from the basic parser to find for structures.
    Only 'top-level' for structures will be found!
    """

    def __init__(self):
        self.out = []

    def visit_For(self, node):
        tv = Parser.get_target_vars(node.target)
        self.out.append(
            {
                "node": node,
                "iter": node.iter,
                "body": {"node": node.body, "target_vars": tv},
                "orelse": {"node": node.orelse, "target_vars": tv},
                "target": node.target,
                "_target_vars": tv,
            }
        )


class ClassDefParser(Parser):
    """Find class definitions
    """

    def __init__(self):
        self.out = {}

    def visit_ClassDef(self, node):
        self.out[node.name] = {
            "node": node,
            "bases": [{"node": node} for node in node.bases],
            "body": node.body,
        }


class FunctionDefParser(Parser):
    """Find function definitions

    A parser which inherits from the basic parser to find function definitions.
    Only 'top-level' for structures will be found!
    """

    def __init__(self):
        self.out = {}

    def visit_FunctionDef(self, node):
        self.out[node.name] = self.parse_node(node)

    @classmethod
    def parse_node(cls, node):
        normal_args = cls.get_arg_tuples(node.args.args, node.args.defaults)
        kwonlyargs = cls.get_arg_tuples(node.args.kwonlyargs, node.args.kw_defaults)
        # TODO: all single args should be tuples like this
        vararg = cls.get_arg(node.args.vararg)
        kwarg = cls.get_arg(node.args.kwarg)
        # create context variables
        target_vars = [arg[0] for arg in normal_args]
        if vararg:
            target_vars.append(vararg)
        if kwarg:
            target_vars.append(kwarg)

        args = cls.get_arg_parts(node.args.args, node.args.defaults, "arg")
        kw_args = cls.get_arg_parts(
            node.args.kwonlyargs, node.args.kw_defaults, "kwonly"
        )
        varargs = cls.get_arg_part(node.args.vararg, None, "vararg")
        kwargs = cls.get_arg_part(node.args.kwarg, None, "kwarg")
        all_args = [*args, varargs, *kw_args, kwargs]

        if isinstance(node, ast.Lambda):
            body_node = node.body
        else:
            bodyMod = wrap_in_module(node.body)
            body_node = FunctionBodyTransformer().visit(bodyMod)

        return {
            "node": node,
            "name": getattr(node, "name", None),
            "args": IndexedDict([(p["name"], p) for p in all_args if p is not None]),
            # TODO: arg is the node counterpart to target_vars
            "_spec1_args": args,
            "*args": varargs,
            "**kwargs": kwargs,
            "body": {"node": body_node, "target_vars": TargetVars(target_vars)},
        }


class LambdaFunctionParser(Parser):
    """Find lambda functions

    A parser which inherits from the basic parser to find lambda functions.
    """

    def __init__(self):
        self.out = []

    def visit_Assign(self, node):
        self.visit(node.value)

    def visit_AugAssign(self, node):
        self.visit(node.value)

    def visit_Call(self, node):
        self.visit(node.func)
        for arg in node.args:
            self.visit(arg)
        for key in node.keywords:
            self.visit(key.value)

    def visit_Lambda(self, node):
        self.out.append(FunctionDefParser.parse_node(node))


class CompParser(Parser):
    def __init__(self):
        self.out = []

    def visit_Assign(self, node):
        self.visit(node.value)

    def visit_AugAssign(self, node):
        self.visit(node.value)

    def build_comp(self, node):
        target = node.generators[0].target
        tv = Parser.get_target_vars(target)
        ifs = node.generators[0].ifs
        self.out.append(
            {
                "node": node,
                "body": {"node": node.elt, "target_vars": tv},
                "target": target,
                "iter": node.generators[0].iter,
                "ifs": [{"node": ifnode, "target_vars": tv} for ifnode in ifs],
                # TODO: 'private' _target_vars, since it shouldn't be set when selecting node,
                #       see remarks in test_list_comp on rewriting
                "_target_vars": tv,
            }
        )


class ListCompParser(CompParser):
    """Find list comprehensions

    A parser which inherits from the CompParser to find list comprehensions.
    """

    def visit_ListComp(self, node):
        self.build_comp(node)


class GeneratorExpParser(CompParser):
    """Find generator expressions

    A parser which inherits from the CompParser to find generator expressions.
    """

    def visit_GeneratorExp(self, node):
        self.build_comp(node)


class DictCompParser(CompParser):
    """Find dictionary comprehensions

    A parser which inherits from the CompParser to find dict comprehensions.
    """

    def visit_DictComp(self, node):
        target = node.generators[0].target
        tv = Parser.get_target_vars(target)
        ifs = node.generators[0].ifs
        self.out.append(
            {
                "node": node,
                "key": {"node": node.key, "target_vars": tv},
                "value": {"node": node.value, "target_vars": tv},
                "target": target,
                "iter": node.generators[0].iter,
                "ifs": [{"node": ifnode, "target_vars": tv} for ifnode in ifs],
                # TODO: 'private' _target_vars, since it shouldn't be set when selecting node,
                #       see remarks in test_list_comp on rewriting
                "_target_vars": tv,
            }
        )


class FunctionBodyTransformer(ast.NodeTransformer):
    # TODO this does not automatically contain line_end information!
    def visit_Nonlocal(self, node):
        new_node = ast.copy_location(ast.Global(names=node.names), node)
        return FunctionBodyTransformer.decorate(new_node, node)

    def visit_Return(self, node):
        new_node = ast.copy_location(ast.Expr(value=node.value), node)
        return FunctionBodyTransformer.decorate(new_node, node)

    @staticmethod
    def decorate(new_node, node):
        new_node.first_token = node.first_token
        new_node.last_token = node.last_token
        return new_node


class WithParser(Parser):
    def __init__(self):
        self.out = []

    def visit_With(self, node):
        items = node.items
        context = [
            {
                "node": item.context_expr,
                "target_vars": self.get_target_vars(item.optional_vars),
                "with_items": item,
                "highlight": node,
            }
            for item in items
        ]

        tv_all = TargetVars(sum([list(c["target_vars"].items()) for c in context], []))

        self.out.append(
            {
                "context": context,
                "body": {"node": node.body, "with_items": items},
                "node": node,
                "n_vars": len(items),
            }
        )


class TryExceptParser(Parser):
    def __init__(self):
        self.out = []

    def visit_Try(self, node):
        handlers = {}

        for handler in node.handlers:
            if isinstance(handler.type, ast.Tuple):
                # of form -- except TypeError, KeyError
                for el in handler.type.elts:
                    handlers[el.id] = self.parse_handler(handler)
            else:
                # either general handler, or single error handler
                k = "all" if not handler.type else self.get_identifier(handler.type)
                handlers[k] = self.parse_handler(handler)

        self.out.append(
            {
                "node": node,
                "body": node.body,
                "orelse": node.orelse or None,
                "finalbody": node.finalbody or None,
                "handlers": handlers,
            }
        )

    @staticmethod
    def get_identifier(obj):
        # if only name needs to be correct
        return getattr(obj, "id", None) or getattr(obj, "attr")
        # if full attr chain must match
        # name = ""
        # while hasattr(obj, "attr"):
        #     name = "." + obj.attr + name
        #     obj = obj.value
        # name = obj.id + name
        # return name

    @staticmethod
    def parse_handler(handler):
        return {"node": handler.body, "target_vars": TargetVars([handler.name])}


parser_dict = {
    "object_accesses": ObjectAccessParser,
    "object_assignments": ObjectAssignmentParser,
    # "operators": OperatorParser,
    "imports": ImportParser,
    "if_elses": IfParser,
    "if_exps": IfExpParser,
    "whiles": WhileParser,
    "for_loops": ForParser,
    "class_defs": ClassDefParser,
    "function_defs": FunctionDefParser,
    "lambda_functions": LambdaFunctionParser,
    "list_comps": ListCompParser,
    "dict_comps": DictCompParser,
    "generator_exps": GeneratorExpParser,
    "withs": WithParser,
    "try_excepts": TryExceptParser,
    "function_calls": FunctionParser,
}



from tcs_pythonwhat.sct_syntax import link_to_state
from tcs_pythonwhat.tasks import getColumnsInProcess
from tcs_pythonwhat.checks.check_object import check_object, check_df, check_keys
from tcs_pythonwhat.checks.has_funcs import has_equal_value

# this is done by the chain for v2
# it's only needed when a new state is created and (possibly) used elsewhere
check_object = link_to_state(check_object)
check_df = link_to_state(check_df)
check_keys = link_to_state(check_keys)


def test_object(
    state,
    name,
    eq_condition="equal",
    eq_fun=None,
    do_eval=True,
    undefined_msg=None,
    incorrect_msg=None,
):

    expand_msg = "" if undefined_msg or incorrect_msg else None
    child = check_object(state, name, undefined_msg, expand_msg=expand_msg)

    if do_eval:
        has_equal_value(child, incorrect_msg)


def test_data_frame(
    state,
    name,
    columns=None,
    undefined_msg=None,
    not_data_frame_msg=None,
    undefined_cols_msg=None,
    incorrect_msg=None,
):
    """Test a pandas dataframe.
    """

    expand_msg = (
        ""
        if undefined_msg or not_data_frame_msg or undefined_cols_msg or incorrect_msg
        else None
    )

    child = check_df(
        state,
        name,
        undefined_msg,
        not_instance_msg=not_data_frame_msg,
        expand_msg=expand_msg,
    )

    # if columns not set, figure them out from solution
    if columns is None:
        columns = getColumnsInProcess(name, child.solution_process)

    for col in columns:
        colstate = check_keys(child, col, missing_msg=undefined_cols_msg)
        has_equal_value(colstate, incorrect_msg=incorrect_msg)



import asttokens

from functools import partialmethod
from collections.abc import Mapping

from protowhat.failure import debugger
from protowhat.Feedback import FeedbackComponent
from protowhat.selectors import DispatcherInterface
from protowhat.State import State as ProtoState
from protowhat.utils import parameters_attr
from tcs_pythonwhat import signatures
from tcs_pythonwhat.converters import get_manual_converters
from tcs_pythonwhat.feedback import Feedback
from tcs_pythonwhat.parsing import (
    TargetVars,
    FunctionParser,
    ObjectAccessParser,
    parser_dict,
)
from tcs_pythonwhat.utils_ast import wrap_in_module


class Context(Mapping):
    def __init__(self, context=None, prev=None):
        self.context = context if context else TargetVars()
        self.prev = prev if prev else {}

        self._items = {**self.prev, **self.context.defined_items()}

    def update_ctx(self, new_ctx):
        return self.__class__(new_ctx, self._items)

    def __getitem__(self, x):
        return self._items[x]

    def __iter__(self):
        return iter(self._items)

    def __len__(self):
        return len(self._items)


@parameters_attr
class State(ProtoState):
    """State of the SCT environment.

    This class holds all information relevevant to test the correctness of an exercise.
    It is coded suboptimally and it will be refactored soon, and documented thouroughly
    after that.

    kwargs:
    ...
     - reporter

    """

    feedback_cls = Feedback

    def __init__(
        self,
        student_code,
        solution_code,
        pre_exercise_code,
        student_process,
        solution_process,
        raw_student_output,
        # solution output
        reporter,
        force_diagnose=False,
        highlight=None,
        highlight_offset=None,
        highlighting_disabled=None,
        feedback_context=None,
        creator=None,
        student_ast=None,
        solution_ast=None,
        student_ast_tokens=None,
        solution_ast_tokens=None,
        student_parts=None,
        solution_parts=None,
        student_context=Context(),
        solution_context=Context(),
        student_env=Context(),
        solution_env=Context(),
    ):
        args = locals().copy()
        self.debug = False

        for k, v in args.items():
            if k != "self":
                setattr(self, k, v)

        self.ast_dispatcher = self.get_dispatcher()

        # Parse solution and student code
        # if possible, not done yet and wanted (ast arguments not False)
        if isinstance(self.student_code, str) and student_ast is None:
            self.student_ast = self.parse(student_code)
        if isinstance(self.solution_code, str) and solution_ast is None:
            with debugger(self):
                self.solution_ast = self.parse(solution_code)

        if highlight is None:  # todo: check parent_state? (move check to reporting?)
            self.highlight = self.student_ast

        self.converters = get_manual_converters()  # accessed only from root state

        self.manual_sigs = None

    def get_manual_sigs(self):
        if self.manual_sigs is None:
            self.manual_sigs = signatures.get_manual_sigs()

        return self.manual_sigs

    def to_child(self, append_message=None, node_name="", **kwargs):
        """Dive into nested tree.

        Set the current state as a state with a subtree of this syntax tree as
        student tree and solution tree. This is necessary when testing if statements or
        for loops for example.
        """
        bad_parameters = set(kwargs) - set(self.parameters)
        if bad_parameters:
            raise ValueError(
                "Недопустимые параметры инициализации для State: %s" % ", ".join(bad_parameters)
            )

        base_kwargs = {
            attr: getattr(self, attr)
            for attr in self.parameters
            if hasattr(self, attr) and attr not in ["ast_dispatcher", "highlight"]
        }

        if append_message and not isinstance(append_message, FeedbackComponent):
            append_message = FeedbackComponent(append_message)
        kwargs["feedback_context"] = append_message
        kwargs["creator"] = {"type": "to_child", "args": {"state": self}}

        def update_kwarg(name, func):
            kwargs[name] = func(kwargs[name])

        def update_context(name):
            update_kwarg(name, getattr(self, name).update_ctx)

        for ast_arg in ["student_ast", "solution_ast"]:
            if isinstance(kwargs.get(ast_arg), list):
                update_kwarg(ast_arg, wrap_in_module)

        if kwargs.get("student_ast") and kwargs.get("student_code") is None:
            kwargs["student_code"] = self.student_ast_tokens.get_text(
                kwargs["student_ast"]
            )
        if kwargs.get("solution_ast") and kwargs.get("solution_code") is None:
            kwargs["solution_code"] = self.solution_ast_tokens.get_text(
                kwargs["solution_ast"]
            )

        for context in [
            "student_context",
            "solution_context",
            "student_env",
            "solution_env",
        ]:
            if context in kwargs:
                if kwargs[context] is not None:
                    update_context(context)
                else:
                    kwargs.pop(context)

        klass = self.SUBCLASSES[node_name] if node_name else State
        init_kwargs = {**base_kwargs, **kwargs}
        child = klass(**init_kwargs)

        extra_attrs = set(vars(self)) - set(self.parameters)
        for attr in extra_attrs:
            # don't copy attrs set on new instances in init
            # the cached manual_sigs is passed
            if attr not in {"ast_dispatcher", "converters"}:
                setattr(child, attr, getattr(self, attr))

        return child

    def has_different_processes(self):
        # process classes have an _identity field that is a tuple
        try:
            return (
                self.student_process._identity[0] != self.solution_process._identity[0]
            )
        except:
            # play it safe (most common)
            return True

    def assert_execution_root(self, fun, extra_msg=""):
        if not (self.is_root or self.is_creator_type("run")):
            with debugger(self):
                self.report(
                    "`%s()` следует вызывать только для полной цепочки, следующей за `Ex()` или `run()`. %s"
                    % (fun, extra_msg)
                )

    def is_creator_type(self, type):
        return self.creator and self.creator.get("type") == type

    def assert_is(self, klasses, fun, prev_fun):
        if self.__class__.__name__ not in klasses:
            with debugger(self):
                self.report(
                    "`%s()` может быть вызван только для  %s."
                    % (fun, " or ".join(["`%s()`" % pf for pf in prev_fun]))
                )

    def assert_is_not(self, klasses, fun, prev_fun):
        if self.__class__.__name__ in klasses:
            with debugger(self):
                self.report(
                    "`%s()` может быть вызван только для %s."
                    % (fun, " or ".join(["`%s()`" % pf for pf in prev_fun]))
                )

    def parse_external(self, code):
        res = (None, None)
        try:
            return self.ast_dispatcher.parse(code)
        except IndentationError as e:
            e.filename = "script.py"
            # no line info for now
            self.report(
                "Ваш код не может быть распаршен из-за ошибки в отступе:<br>`%s.`"
                % str(e)
            )

        except SyntaxError as e:
            e.filename = "script.py"
            # no line info for now
            self.report(
                "Ваш код не может быть выполнен из-за синтаксической ошибки:<br>`%s.`" % str(e)
            )

        # Can happen, can't catch this earlier because we can't differentiate between
        # TypeError in parsing or TypeError within code (at runtime).
        except:
            self.report("Что-то пошло не так при парсинге вашего кода.")

        return res

    def parse_internal(self, code):
        try:
            return self.ast_dispatcher.parse(code)
        except Exception as e:
            self.report(
                "Что-то пошло не так при парсинге solution-кода: %s" % str(e)
            )

    def parse(self, text):
        if self.debug:
            parse_method = self.parse_internal
            token_attr = "solution_ast_tokens"
        else:
            parse_method = self.parse_external
            token_attr = "student_ast_tokens"

        tokens, ast = parse_method(text)
        setattr(self, token_attr, tokens)

        return ast

    def get_dispatcher(self):
        try:
            return Dispatcher(self.pre_exercise_code)
        except Exception as e:
            with debugger(self):
                self.report("Что-то пошло не так при парсинге PEC: %s" % str(e))


class Dispatcher(DispatcherInterface):
    _context_cache = dict()

    def __init__(self, context_code=""):
        self._parser_cache = dict()
        context_ast = getattr(self._context_cache, context_code, None)
        if context_ast is None:
            context_ast = self._context_cache[context_code] = self.parse(context_code)[
                1
            ]
        self.context_mappings = self._getx(FunctionParser, "mappings", context_ast)

    def find(self, name, node, *args, **kwargs):
        return getattr(self, name)(node)

    def parse(self, code):
        res = asttokens.ASTTokens(code, parse=True)
        return res, res.tree

    # add methods for retrieving parser outputs --------------------------
    def _getx(self, Parser, ext_attr, tree):
        """getter for Parser outputs"""
        # return cached output if possible
        cache_key = Parser.__name__ + str(hash(tree))
        if self._parser_cache.get(cache_key):
            p = self._parser_cache[cache_key]
        else:
            # otherwise, run parser over tree
            p = Parser()
            # set mappings for parsers that inspect attribute access
            if ext_attr != "mappings" and Parser in [
                FunctionParser,
                ObjectAccessParser,
            ]:
                p.mappings = self.context_mappings.copy()
            # run parser
            p.visit(tree)
            # cache
            self._parser_cache[cache_key] = p
        return getattr(p, ext_attr)


# put a function on the dispatcher
for k, Parser in parser_dict.items():
    setattr(Dispatcher, k, partialmethod(Dispatcher._getx, Parser, "out"))

# mappings from ObjectAccessParser
prop_oa_map = partialmethod(Dispatcher._getx, ObjectAccessParser, "mappings")
setattr(Dispatcher, "oa_mappings", prop_oa_map)

# mappings from FunctionParser
prop_map = partialmethod(Dispatcher._getx, FunctionParser, "mappings")
setattr(Dispatcher, "mappings", prop_map)


# State subclasses based on parsed output -------------------------------------
State.SUBCLASSES = {
    node_name: type(node_name, (State,), {}) for node_name in parser_dict
}


# global setters on State -----------------------------------------------------
def set_converter(key, fundef):
    # note that root state is set on the State class in test_exercise
    State.root_state.converters[key] = fundef



import inspect
from inspect import Parameter as param
import tcs_pythonwhat
from tcs_pythonwhat.tasks import getSignatureFromObjInProcess


def sig_from_params(*args):
    return inspect.Signature(list(args))


def sig_from_obj(obj_char):
    return getSignatureFromObjInProcess(
        obj_char, pythonwhat.State.State.root_state.solution_process
    )


def get_manual_sigs():
    manual_sigs = {
        # builtins
        "abs": [param("x", param.POSITIONAL_ONLY)],
        "all": [param("iterable", param.POSITIONAL_ONLY)],
        "any": [param("iterable", param.POSITIONAL_ONLY)],
        "ascii": [param("obj", param.POSITIONAL_ONLY)],
        "bin": [param("number", param.POSITIONAL_ONLY)],
        "bool": [param("x", param.POSITIONAL_OR_KEYWORD)],
        "chr": [param("i", param.POSITIONAL_ONLY)],
        "callable": [param("obj", param.POSITIONAL_ONLY)],
        "classmethod": [param("function", param.POSITIONAL_ONLY)],
        "complex": [
            param("imag", param.POSITIONAL_OR_KEYWORD, default=0),
            param("real", param.POSITIONAL_OR_KEYWORD, default=0),
        ],
        "delattr": [
            param("obj", param.POSITIONAL_ONLY),
            param("name", param.POSITIONAL_ONLY),
        ],
        "dir": [param("object", param.POSITIONAL_OR_KEYWORD, default=None)],
        "divmod": [
            param("x", param.POSITIONAL_ONLY),
            param("y", param.POSITIONAL_ONLY),
        ],
        "enumerate": [
            param("iterable", param.POSITIONAL_ONLY),
            param("start", param.POSITIONAL_OR_KEYWORD, default=0),
        ],
        "float": [param("x", param.POSITIONAL_OR_KEYWORD)],
        "getattr": [
            param("object", param.POSITIONAL_ONLY),
            param("name", param.POSITIONAL_ONLY),
            param("default", param.POSITIONAL_ONLY, default=None),
        ],
        "hasattr": [
            param("obj", param.POSITIONAL_ONLY),
            param("name", param.POSITIONAL_ONLY),
        ],
        "hash": [param("obj", param.POSITIONAL_ONLY)],
        "hex": [param("number", param.POSITIONAL_ONLY)],
        "id": [param("obj", param.POSITIONAL_ONLY)],
        "int": [
            param("x", param.POSITIONAL_OR_KEYWORD),
            param("base", param.POSITIONAL_OR_KEYWORD, default=10),
        ],
        "isinstance": [
            param("obj", param.POSITIONAL_ONLY),
            param("class_or_tuple", param.POSITIONAL_ONLY),
        ],
        "issubclass": [
            param("cls", param.POSITIONAL_ONLY),
            param("class_or_tuple", param.POSITIONAL_ONLY),
        ],
        "list": [param("iterable", param.POSITIONAL_ONLY, default=None)],
        "len": [param("obj", param.POSITIONAL_ONLY)],
        "oct": [param("number", param.POSITIONAL_ONLY)],
        "open": [
            param("file", param.POSITIONAL_OR_KEYWORD),
            param("mode", param.POSITIONAL_OR_KEYWORD, default="r"),
            param("buffering", param.POSITIONAL_OR_KEYWORD, default=1),
            param("encoding", param.POSITIONAL_OR_KEYWORD, default=None),
            param("errors", param.POSITIONAL_OR_KEYWORD, default=None),
            param("newline", param.POSITIONAL_OR_KEYWORD, default=None),
            param("closefd", param.POSITIONAL_OR_KEYWORD, default=None),
            param("opener", param.POSITIONAL_OR_KEYWORD, default=None),
        ],
        "ord": [param("c", param.POSITIONAL_ONLY)],
        "pow": [
            param("x", param.POSITIONAL_ONLY),
            param("y", param.POSITIONAL_ONLY),
            param("z", param.POSITIONAL_ONLY, default=None),
        ],
        "print": [param("value", param.POSITIONAL_ONLY)],
        "repr": [param("obj", param.POSITIONAL_ONLY)],
        "reversed": [param("sequence", param.POSITIONAL_ONLY)],
        "round": [
            param("number", param.POSITIONAL_OR_KEYWORD),
            param("ndigits", param.POSITIONAL_OR_KEYWORD, default=0),
        ],
        "set": [param("iterable", param.POSITIONAL_ONLY, default=None)],
        # Difference v3.4 vs v3.5!!!
        "setattr": [
            param("obj", param.POSITIONAL_ONLY),
            param("name", param.POSITIONAL_ONLY),
            param("value", param.POSITIONAL_ONLY),
        ],
        "sorted": [
            param("iterable", param.POSITIONAL_ONLY),
            param("key", param.POSITIONAL_OR_KEYWORD, default=None),
            param("reverse", param.POSITIONAL_OR_KEYWORD, default=False),
        ],
        "str": [param("object", param.POSITIONAL_OR_KEYWORD)],
        "sum": [
            param("iterable", param.POSITIONAL_ONLY),
            param("start", param.POSITIONAL_ONLY, default=0),
        ],
        "tuple": [param("iterable", param.POSITIONAL_ONLY, default=None)],
        "type": [param("object", param.POSITIONAL_ONLY)],
        "vars": [param("object", param.POSITIONAL_ONLY)],
        # int
        # str
        "str.center": [
            param("width", param.POSITIONAL_ONLY),
            param("fillchar", param.POSITIONAL_ONLY, default=" "),
        ],
        # list
        "list.append": [param("object", param.POSITIONAL_ONLY)],
        "list.count": [param("value", param.POSITIONAL_ONLY)],
        # dict
        # numpy
        "numpy.array": [
            param("object", param.POSITIONAL_OR_KEYWORD),
            param("dtype", param.POSITIONAL_OR_KEYWORD, default=None),
            param("copy", param.POSITIONAL_OR_KEYWORD, default=True),
            param("order", param.POSITIONAL_OR_KEYWORD, default=None),
            param("subok", param.POSITIONAL_OR_KEYWORD, default=False),
            param("ndmin", param.POSITIONAL_OR_KEYWORD, default=0),
        ],
        "numpy.random.seed": [param("seed", param.POSITIONAL_OR_KEYWORD, default=None)],
        "numpy.random.rand": [
            param("d0", param.POSITIONAL_ONLY, default=None),
            param("d1", param.POSITIONAL_ONLY, default=None),
            param("d2", param.POSITIONAL_ONLY, default=None),
            param("d3", param.POSITIONAL_ONLY, default=None),
            param("d4", param.POSITIONAL_ONLY, default=None),
            param("d5", param.POSITIONAL_ONLY, default=None),
            param("d6", param.POSITIONAL_ONLY, default=None),
        ],
        "numpy.random.randint": [
            param("low", param.POSITIONAL_OR_KEYWORD),
            param("high", param.POSITIONAL_OR_KEYWORD, default=None),
            param("size", param.POSITIONAL_OR_KEYWORD, default=None),
            param("dtype", param.POSITIONAL_OR_KEYWORD, default="l"),
        ],
        "numpy.random.choice": [
            param("a", param.POSITIONAL_OR_KEYWORD),
            param("size", param.POSITIONAL_OR_KEYWORD, default=None),
            param("replace", param.POSITIONAL_OR_KEYWORD, default=True),
            param("p", param.POSITIONAL_OR_KEYWORD, default=None),
        ],
        "numpy.random.normal": [
            param("loc", param.POSITIONAL_OR_KEYWORD, default=0.0),
            param("scale", param.POSITIONAL_OR_KEYWORD, default=1.0),
            param("size", param.POSITIONAL_OR_KEYWORD, default=None),
        ],
        "numpy.random.poisson": [
            param("lam", param.POSITIONAL_OR_KEYWORD, default=1.0),
            param("size", param.POSITIONAL_OR_KEYWORD, default=None),
        ],
        "numpy.random.binomial": [
            param("n", param.POSITIONAL_OR_KEYWORD),
            param("p", param.POSITIONAL_OR_KEYWORD),
            param("size", param.POSITIONAL_OR_KEYWORD, default=None),
        ],
        "numpy.random.shuffle": [param("x", param.POSITIONAL_OR_KEYWORD)],
        "numpy.random.permutation": [param("x", param.POSITIONAL_OR_KEYWORD)],
        # others
        "math.radians": [param("x", param.POSITIONAL_ONLY)],
    }
    return manual_sigs



import pprint as pp
import itertools
import inspect
from functools import partial
from collections import OrderedDict
from tcs_pythonwhat import test_funcs
from tcs_pythonwhat.State import State
from tcs_pythonwhat.checks.check_wrappers import state_partial

TEST_NAMES = [
    "test_mc",
    "test_or",
    "test_with",
    "test_import",
    "test_object",
    "test_correct",
    "test_if_else",
    "test_for_loop",
    "test_function",
    "test_list_comp",
    "test_function_v2",
    "test_data_frame",
    "test_while_loop",
    "test_student_typed",
    "test_object_accessed",
    "test_output_contains",
    "test_expression_result",
    "test_expression_output",
    "test_function_definition",
    "test_object_after_expression",
]

SUB_TESTS = {
    "test_if_else": ["test", "body", "orelse"],
    "test_list_comp": ["comp_iter", "body", "ifs"],
    "check_correct": ["check", "diagnose"],
    "test_for_loop": ["for_iter", "body", "orelse"],
    "test_while_loop": ["test", "body", "orelse"],
    "test_with": ["context_tests", "body"],
    "test_function_definition": ["body"],
    "check_or": ["tests"],
}


class Tree(object):
    def __init__(self):
        """
        Represent a tree of nodes, by holding the currently active node.

        This class is necessary to put sub-tests onto a graph, because they may
        exist inside of function calls. By getting the currently active node
        from the tree, Probe instances may add subtests as children on that node,
        and update it when recursing over sub-tests.
        
        """
        self.root = Node(name="root")
        self.crnt_node = self.root

    @classmethod
    def str_branch(cls, node, str_func=lambda s: str(s)):
        # dict(getattr(s.data.get("bound_args", {}), 'arguments', {}))
        f = node.data.get("func")
        this_node = (
            "  " * node.depth
            + "("
            + getattr(f, "__name__", node.name)
            + str_func(node)
            + ")\n"
        )
        return this_node + "".join(
            map(lambda x: cls.str_branch(x, str_func), node.child_list)
        )

    def __str__(self):
        return self.str_branch(self.crnt_node)

    def descend(self, node=None):
        node = self.crnt_node if node is None else node
        children = map(self.descend, node.child_list)
        base = [node] if node.name != "root" else []
        return sum(children, base)

    def __iter__(self):
        for ii in self.descend(self.crnt_node):
            yield ii


class Node(object):
    def __init__(self, child_list=None, data=None, name="unnamed", arg_name=""):
        """
        Hold a function call with its bound arguments, along with child nodes.
        
        """
        self.parent = None
        self.name = name
        self.arg_name = arg_name
        self.child_list = [] if child_list is None else child_list
        self.data = {} if data is None else data
        self.updated = False
        # hacky way to add their argument name when a function of tests was given

    def __call__(self, state=None):
        """Call original function with its arguments, and optional state"""
        ba = self.data["bound_args"]
        if state:
            func = self.data["func"]
            ba = inspect.signature(func).bind(state, *ba.args[1:], **ba.kwargs)
            func(*ba.args, **ba.kwargs)
            return state
        else:
            self.data["func"](*ba.args, **ba.kwargs)
            ba.apply_defaults()
            return ba.arguments["state"]

    def __str__(self):
        return pp.pformat(
            dict(getattr(self.data.get("bound_args", {}), "arguments", {}))
        )

    def __iter__(self):
        for c in self.child_list:
            yield c

    def partial(self):
        """Return partial of original function call"""
        ba = self.data["bound_args"]
        return state_partial(self.data["func"], *ba.args[1:], **ba.kwargs)

    def update_child_calls(self):
        """Replace child nodes on original function call with their partials"""

        for node in filter(lambda n: len(n.arg_name), self.child_list):
            self.data["bound_args"].arguments[node.arg_name] = node.partial()
        self.updated = True

    def remove_child(self, node):
        index = self.child_list.index(node)
        del self.child_list[index]
        return index

    def add_child(self, child):
        # since it is a tree, there is only one parent
        # note this means we do not allow edges between same layer units
        if child.parent:
            child.parent.remove_child(child)
        child.parent = self
        self.child_list.append(child)

    def descend(self, include_me=True):
        """Descend depth first into all child nodes"""
        if include_me:
            yield self

        for child in self.child_list:
            yield child
            yield from child.descend()

    @property
    def depth(self):
        if self.parent:
            return self.parent.depth + 1
        else:
            return 0


class NodeList(Node):
    def partial(self):
        return [node.partial() for node in self.child_list]

    def update_child_calls(self):
        pass


class Probe(object):
    def __init__(self, tree, f, eval_on_call=False):
        self.tree = tree
        self.f = f
        self.test_name = f.__name__
        self.eval_on_call = eval_on_call
        # TODO: auto sub_test detection
        self.sub_tests = SUB_TESTS.get(self.test_name) or []

    def __call__(self, *args, **kwargs):
        """Bind arguments to original function signature, and store in a Node

        This is used to discover what tests and sub-tests the SCT would like to 
        call, and defer them for later execution via their node instance. Node
        instances are assembled into a tree.

        """
        if (len(args) > 0 and not isinstance(args[0], State)) or len(args) == 0:
            # no state placeholder if a state is passed
            args = ["state_placeholder"] + list(args)

        bound_args = inspect.signature(self.f).bind(*args, **kwargs)

        data = dict(bound_args=bound_args, func=self.f)
        this_node = Node(data=data, name=self.test_name)
        if self.tree is not None:
            self.tree.crnt_node.add_child(this_node)

        # First pass to set up branches off node
        arguments = bound_args.arguments
        for subtest in self.sub_tests:  # TODO: auto sub test detection
            if subtest in arguments and arguments[subtest]:
                self.build_sub_test_nodes(
                    arguments[subtest], self.tree, this_node, subtest
                )

        # Second pass to build node and all its children into a subtest
        for node in this_node.descend(include_me=True):
            if node.updated:  # already built, e.g. node used multiple times
                continue
            else:
                node.update_child_calls()

        if self.eval_on_call:
            return this_node()
        else:
            return this_node

    @staticmethod
    def build_sub_test_nodes(test, tree, node, arg_name):
        # note that I've made the strong assumption that
        # if not a function, then test is a dict, list or tuple of them
        if isinstance(test, (list, tuple)):
            nl = NodeList(name="List", arg_name=arg_name)
            node.add_child(nl)
            for ii, f in enumerate(test):
                Probe.build_sub_test_nodes(f, tree, nl, str(ii))
        elif isinstance(test, Node):
            # test was a lambdaless subtest call, which produced a node
            # so need to tell it what its arg_name was on parent test
            test.arg_name = arg_name
            node.add_child(test)
        elif callable(test):
            # test was inside a lambda, function containing subtests or v2 F() chain object with subtests
            # since either may contain multiple subtests, we put them in a node list
            nl = NodeList(name="ListDeferred", arg_name=arg_name)
            node.add_child(nl)
            if tree is not None:
                prev_node, tree.crnt_node = tree.crnt_node, nl
                test()
                tree.crnt_node = prev_node
            else:
                test()
        elif test is not None:
            raise Exception("Ожидалась функция или список/кортеж/словрь функций")


def build_probe_context():
    tree = Tree()
    probe_context = {s: Probe(tree, getattr(test_funcs, s)) for s in TEST_NAMES}
    return tree, probe_context



#!/usr/bin/env python

import re
import ast
from os import path
from setuptools import setup

PACKAGE_NAME = "tcs_pythonwhat"
REQUIREMENT_NAMES = ["protowhat", "markdown2", "jinja2", "asttokens", "dill"]
PEER_REQUIREMENTS = ["numpy", "pandas"]

HERE = path.abspath(path.dirname(__file__))
VERSION_FILE = path.join(HERE, PACKAGE_NAME, "__init__.py")
REQUIREMENTS_FILE = path.join(HERE, "requirements.txt")
README_FILE = path.join(HERE, "README.md")

with open(VERSION_FILE, encoding="utf-8") as fp:
    _version_re = re.compile(r"__version__\s+=\s+(.*)")
    VERSION = str(ast.literal_eval(_version_re.search(fp.read()).group(1)))
with open(REQUIREMENTS_FILE, encoding="utf-8") as fp:
    req_txt = fp.read()
    _requirements_re_template = r"^({}(?:\s*[~<>=]+\s*\S*)?)\s*(?:#.*)?$"
    REQUIREMENTS = [
        re.search(_requirements_re_template.format(requirement), req_txt, re.M).group(0)
        for requirement in REQUIREMENT_NAMES
    ] + PEER_REQUIREMENTS
with open(README_FILE, encoding="utf-8") as fp:
    README = fp.read()

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    packages=[PACKAGE_NAME, "tcs_pythonwhat.checks", "tcs_pythonwhat.test_funcs"],
    install_requires=REQUIREMENTS,
    description="Submission correctness tests for Python",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Filip Schouwenaars",
    author_email="filip@datacamp.com",
    maintainer="Jeroen Hermans",
    maintainer_email="content-engineering@datacamp.com",
    url="https://github.com/datacamp/pythonwhat",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
)



from protowhat.sct_syntax import EagerChain, ExGen, LazyChainStart, state_dec_gen, LazyChain
from tcs_pythonwhat.checks.check_wrappers import scts
from tcs_pythonwhat.State import State
from tcs_pythonwhat.probe import Node, Probe, TEST_NAMES
from tcs_pythonwhat.utils import include_v1
from tcs_pythonwhat import test_funcs
from functools import wraps

# TODO: could define scts for check_wrappers at the module level
sct_dict = scts.copy()


def multi_dec(f):
    """Decorator for multi to remove nodes for original test functions from root node"""

    @wraps(f)
    def wrapper(*args, **kwargs):
        args = (
            args[0] if len(args) == 1 and isinstance(args[0], (list, tuple)) else args
        )
        for arg in args:
            if isinstance(arg, Node) and arg.parent.name is "root":
                arg.parent.remove_child(arg)
                arg.update_child_calls()
        return f(*args, **kwargs)

    return wrapper


state_dec = state_dec_gen(sct_dict)

# todo: __all__?
assert ExGen
assert LazyChainStart


def Ex(state=None):
    return EagerChain(state=state or State.root_state, chainable_functions=sct_dict)


def F():
    return LazyChain(chainable_functions=sct_dict)


def get_chains():
    return {
        "Ex": ExGen(sct_dict, State.root_state),
        "F": LazyChainStart(sct_dict),
    }


if include_v1():
    # Prepare SCTs that may be chained attributes ----------------------
    # decorate functions that may try to run test_* function nodes as subtests
    # so they remove those nodes from the tree
    for k in ["multi", "with_context"]:
        sct_dict[k] = multi_dec(sct_dict[k])

    # allow test_* functions as chained attributes
    for k in TEST_NAMES:
        sct_dict[k] = Probe(tree=None, f=getattr(test_funcs, k), eval_on_call=True)

    # original logical test_* functions behave like multi
    # this is necessary to allow them to take check_* funcs as args
    # since probe behavior will try to call all SCTs passed (assuming they're also probes)
    for k in ["test_or", "test_correct"]:
        sct_dict[k] = multi_dec(getattr(test_funcs, k))

# Prepare check_funcs to be used alone (e.g. test = check_with().check_body())
v2_check_functions = {k: state_dec(v) for k, v in scts.items()}



import inspect
import tcs_pythonwhat


def get_manual_converters():

    converters = {
        "pandas.io.excel.ExcelFile": lambda x: x.io,
        "pandas.io.excel._base.ExcelFile": lambda x: x.io,
        "builtins.dict_keys": lambda x: sorted(x),
        "builtins.dict_items": lambda x: sorted(x),
        "bs4.BeautifulSoup": lambda x: str(x),
        "bs4.element.Tag": lambda x: str(x),
        "bs4.element.NavigableString": lambda x: str(x),
        "bs4.element.ResultSet": lambda x: [str(res) for res in x],
        "h5py._hl.files.File": lambda x: x.file.filename,
        "h5py._hl.group.Group": lambda x: x.file.filename
        + "_"
        + str([x for x in x.keys()]),
        "sqlalchemy.engine.base.Engine": lambda x: x.url.database,
    }

    return converters



import re
import numpy as np
import pandas as pd
from pandas.testing import assert_frame_equal, assert_series_equal
from tcs_pythonwhat.tasks import *
from protowhat.Test import Test

"""
This file contains all tests that can be done on specific objects. All tests are represented as
an object. Tests that are alike can inherit from the same superclass. A test is first initialized
and can then be performed by calling the 'test()' function. The result will be stored inside
the result boolean. A test contains a failure message, which can be used by the reporter to
show when the test failed.
"""


# Testing definition


class DefinedProcessTest(Test):
    def __init__(self, name, process, feedback):
        super().__init__(feedback)
        self.name = name
        self.process = process

    def test(self):
        self.result = isDefinedInProcess(self.name, self.process)


class DefinedCollTest(Test):
    """
    Check if an object with a certain name is defined in a collection.

    Attributes:
        feedback (str): A string containing the failure message in case the test fails.
        name (str): Contains the name of the object that is searched for.
        collection (list/dict/set): Contains any object on which the 'in' operator can be performed.
        result (bool): True if the test succeed, False if it failed. None if it hasn't been tested yet.
    """

    def __init__(self, name, collection, feedback):
        super().__init__(feedback)
        self.name = name
        self.collection = collection

    def test(self):
        self.result = self.name in self.collection


class DefinedCollProcessTest(Test):
    def __init__(self, name, key, process, feedback):
        super().__init__(feedback)
        self.name = name
        self.key = key
        self.process = process

    def test(self):
        self.result = isDefinedCollInProcess(self.name, self.key, self.process)


# Testing class


class InstanceProcessTest(Test):
    def __init__(self, name, klass, process, feedback):
        super().__init__(feedback)
        self.name = name
        self.klass = klass
        self.process = process

    def test(self):
        self.result = isInstanceInProcess(self.name, self.klass, self.process)


# Testing equality


class EqualTest(Test):
    """
    Check if two objects are equal. Equal means the objects are exactly the same.
    This test should only be used with numeric variables (for now).

    Attributes:
        feedback (str): A string containing the failure message in case the test fails.
        obj1 (str): The first object that should be compared with.
        obj2 (str): This object is compared to obj1.
        result (bool): True if the test succeed, False if it failed. None if it hasn't been tested yet.
    """

    def __init__(self, obj1, obj2, feedback, func=None):
        super().__init__(feedback)
        self.obj1 = obj1
        self.obj2 = obj2
        self.func = func if func is not None else is_equal

    def test(self):
        """
        Perform the actual test. result is set to False if the objects differ, True otherwise.
        """
        self.result = np.array(self.func(self.obj1, self.obj2)).all()


# Helpers for testing equality


def areinstance(x, y, tuple_of_classes):
    return isinstance(x, tuple_of_classes) and isinstance(y, tuple_of_classes)


# For equality of ndarrays, list, dicts, pd Series and pd DataFrames:
# First try to the faster equality functions. If these don't pass,
# Run the assertions that are typically slower.
def is_equal(x, y):
    try:
        if areinstance(x, y, (Exception,)):
            # Types of errors don't matter (this is debatable)
            return str(x) == str(y)
        if areinstance(x, y, (np.ndarray, dict, list, tuple)):
            np.testing.assert_equal(x, y)
            return True
        elif areinstance(x, y, (map, filter)):
            return np.array_equal(list(x), list(y))
        elif areinstance(x, y, (pd.DataFrame,)):
            if x.equals(y):
                return True
            assert_frame_equal(x, y)
            return True
        elif areinstance(x, y, (pd.Series,)):
            if x.equals(y):
                return True
            assert_series_equal(x, y)
            return True
        else:
            return x == y

    except Exception:
        return False


# Others


class BiggerTest(EqualTest):
    """
    Check if the first object is greater than another.
    """

    def __init__(self, *args):
        super().__init__(*args, func=lambda obj1, obj2: obj1 > obj2)


class StringContainsTest(Test):
    """
    Check if a string is present in a text. Can use literal strings or a regex.

    Attributes:
        feedback (str): A string containing the failure message in case the test fails.
        string (regex/str):  String or regular expression which is searched for.
        search_string (str): The text in which is searched.
        pattern (bool): If set to True, string is matched with a regex. Literal otherwise.
        result (bool): True if the test succeed, False if it failed. None if it hasn't been tested yet.
    """

    def __init__(self, string, search_string, pattern, feedback):
        """
        Initialize with a string to look for, a string to search and whether or not to look for a pattern.

        Args:
            string (regex/str):  The string to look for will be set to this.
            search_string (str): The string to search in will be set to this.
            pattern (bool): The pattern boolean will be set to this.
            feedback (str): The failure message will be set to this.
        """
        super().__init__(feedback)
        self.string = string
        self.search_string = search_string
        self.pattern = pattern

    def test(self):
        """
        Perform the actual test. result will be True if string is found (whether or not with a pattern),
        False otherwise.
        """
        if self.pattern:
            self.result = re.search(self.search_string, self.string) is not None
        else:
            self.result = self.string.find(self.search_string) != -1



import ast

from protowhat.failure import debugger


def wrap_in_module(node):
    new_node = ast.Module(node, [])
    if isinstance(node, list):
        if len(node) > 0:
            new_node.first_token = node[0].first_token
            new_node.last_token = node[-1].last_token
        else:
            pass  # do nothing
    else:
        new_node.first_token = node.first_token
        new_node.last_token = node.first_token
    return new_node


def assert_ast(state, element, fmt_kwargs):
    err_msg = (
        "SCT fails on solution: "
        "You are zooming in on the {{part}}, but it is not an AST, so it can't be re-run."
        " If this error occurred because of ``check_args()``,"
        "you may have to refer to your argument differently, e.g. `['args', 0]` or `['kwargs', 'a']`. "
        "Read https://pythonwhat.readthedocs.io/en/latest/articles/checking_function_calls.html#signatures for more info."
    )
    # element can also be { 'node': AST }
    if isinstance(element, dict):
        element = element["node"]
    if isinstance(element, ast.AST):
        return
    if isinstance(element, list) and all([isinstance(el, ast.AST) for el in element]):
        return
    with debugger(state):
        state.report(err_msg, fmt_kwargs)
